@@ -2,6 +2,7 @@
 #define STRICT_R_HEADERS
 
 #include <Rinternals.h>
+#include <Rdefines.h>
 #include <R_ext/Rdynload.h>
 #include <librsvg/rsvg.h>
 #include <cairo.h>
@@ -16,13 +17,43 @@ typedef struct {
   size_t size;
 } memory;
 
+#if LIBRSVG_CHECK_VERSION(2,52,0)
+static int unit_to_px(RsvgLength x){
+  double l = x.length;
+  switch(x.unit){
+  case RSVG_UNIT_PX:
+  case RSVG_UNIT_PT:
+    return l;
+  case RSVG_UNIT_IN:
+    return 300*l;
+  case RSVG_UNIT_CM:
+    return 100*l;
+  default:
+    return 800;
+  }
+}
+#endif
+
+static void setup_render_handle(RsvgHandle *svg, cairo_t *cr){
+#if LIBRSVG_CHECK_VERSION(2,52,0)
+  GError *err = NULL;
+  RsvgRectangle viewport = {0.0};
+  //viewport.width = width;
+  //viewport.height = height;
+  rsvg_handle_render_document(svg, cr, &viewport, &err);
+  if(err != NULL)
+    Rf_error("Failure in rsvg_handle_render_document: %s", err->message);
+#else
+  if(!rsvg_handle_render_cairo(svg, cr))
+    Rf_error("Cairo failed to render svg");
+#endif
+}
 
 static SEXP write_bitmap(RsvgHandle *svg, int width, int height, double sx, double sy){
   cairo_surface_t *canvas = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
   cairo_t *cr = cairo_create(canvas);
   cairo_scale(cr, sx, sy);
-  if(!rsvg_handle_render_cairo(svg, cr))
-    Rf_error("Cairo failed to render svg");
+  setup_render_handle(svg, cr);
   int stride = cairo_image_surface_get_stride(canvas); //should be equal to width * channels
   int size = stride * height;
   cairo_surface_flush(canvas);
@@ -40,6 +71,44 @@ static SEXP write_bitmap(RsvgHandle *svg, int width, int height, double sx, doub
   return image;
 }
 
+static SEXP write_native_raster(RsvgHandle *svg, int width, int height, double sx, double sy){
+  cairo_surface_t *canvas = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
+  cairo_t *cr = cairo_create(canvas);
+  cairo_scale(cr, sx, sy);
+  setup_render_handle(svg, cr);
+  cairo_surface_flush(canvas);
+
+  // Allocate an integer matrix of class 'nativeRaster'
+  SEXP image = PROTECT(Rf_allocVector(INTSXP, width*height));
+  SEXP dim = PROTECT(Rf_allocVector(INTSXP, 2));
+  INTEGER(dim)[0] = height;
+  INTEGER(dim)[1] = width;
+  Rf_setAttrib(image, R_DimSymbol, dim);
+  SET_CLASS(image, Rf_mkString("nativeRaster"));
+
+  // Get a pointer to the raw bytes from cairo in CAIRO_FORMAT_ARGB32
+  unsigned char *bp  = cairo_image_surface_get_data(canvas);
+  int *nrp = INTEGER(image);
+
+  // Note: this is not properly handling stride and is just assuming
+  // that the width of the canvas is width*channels
+  // Twiddle the colour channels from RGBA to ARGB
+  for (int i=0; i < height * width; i++) {
+    nrp[i] =
+      (bp[(i << 2) + 0]) << 16 |   // RGBA -> ARGB
+      (bp[(i << 2) + 1]) <<  8 |   //
+      (bp[(i << 2) + 2]) <<  0 |   //
+      (bp[(i << 2) + 3]) << 24 ;   //
+  }
+
+  // Free and return
+  UNPROTECT(2);
+  g_object_unref(svg);
+  cairo_surface_destroy(canvas);
+  cairo_destroy(cr);
+  return image;
+}
+
 static cairo_status_t write_func(void *ctx, const unsigned char *data, unsigned int length) {
   memory *mem = (memory*) ctx;
   #ifdef _WIN32
@@ -58,8 +127,7 @@ static SEXP write_png(RsvgHandle *svg, int width, int height, double sx, double
   cairo_surface_t *canvas = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
   cairo_t *cr = cairo_create(canvas);
   cairo_scale(cr, sx, sy);
-  if(!rsvg_handle_render_cairo(svg, cr))
-    Rf_error("Cairo failed to render svg");
+  setup_render_handle(svg, cr);
   memory mem = {NULL, 0};
   cairo_surface_write_to_png_stream(canvas, write_func, &mem);
   cairo_surface_flush(canvas);
@@ -72,13 +140,14 @@ static SEXP write_png(RsvgHandle *svg, int width, int height, double sx, double
   return res;
 }
 
-static SEXP write_stream(RsvgHandle *svg, int width, int height, double sx, double sy, cairo_surface_t* (*fun) (cairo_write_func_t, void *, double, double)) {
+static SEXP write_stream(RsvgHandle *svg, int width, int height, double sx, double sy, cairo_surface_t* (*fun) (cairo_write_func_t, void *, double, double), cairo_bool_t is_eps) {
   memory buf = {NULL, 0};
   cairo_surface_t *canvas = fun(write_func, &buf, width, height);
+  if(is_eps)
+    cairo_ps_surface_set_eps(canvas, TRUE);
   cairo_t *cr = cairo_create(canvas);
   cairo_scale(cr, sx, sy);
-  if(!rsvg_handle_render_cairo(svg, cr))
-    Rf_error("Cairo failed to render svg");
+  setup_render_handle(svg, cr);
   cairo_surface_show_page(canvas);
   cairo_surface_flush(canvas);
   cairo_surface_destroy(canvas);
@@ -107,8 +176,24 @@ SEXP R_rsvg(SEXP data, SEXP rwidth, SEXP rheight, SEXP format, SEXP css){
     Rf_warning("An external CSS file was specified but this requires at least librsvg 2.48 (you have %s)", LIBRSVG_VERSION);
 #endif
   }
+#if LIBRSVG_CHECK_VERSION(2,52,0)
+  gboolean has_width;
+  RsvgLength in_width;
+  gboolean has_height;
+  RsvgLength in_height;
+  gboolean has_viewbox;
+  RsvgRectangle viewbox;
+  rsvg_handle_get_intrinsic_dimensions (svg, &has_width, &in_width, &has_height, &in_height, &has_viewbox, &viewbox);
+  double input_width = has_viewbox ? viewbox.width : (has_width ? unit_to_px(in_width) : 800);
+  double input_height = has_viewbox ? viewbox.height : (has_height ? unit_to_px(in_height) : 800);
+  //REprintf("Size: %fx%f\n", input_width, input_height);
+#else
   RsvgDimensionData dimensions;
   rsvg_handle_get_dimensions(svg, &dimensions);
+  int input_width = dimensions.width;
+  int input_height = dimensions.height;
+  //REprintf("Size: %fx%f\n", (double)input_width, (double)input_height);
+#endif
 
   //scale into the requested resolution
   double width;
@@ -116,22 +201,22 @@ SEXP R_rsvg(SEXP data, SEXP rwidth, SEXP rheight, SEXP format, SEXP css){
   double sx;
   double sy;
   if(rwidth == R_NilValue && rheight == R_NilValue){
-    width = dimensions.width;
-    height = dimensions.height;
+    width = input_width;
+    height = input_height;
     sx = sy = 1;
   } else if(rwidth != R_NilValue && rheight != R_NilValue){
     width = Rf_asInteger(rwidth);
     height = Rf_asInteger(rheight);
-    sx = width / dimensions.width;
-    sy = height / dimensions.height;
+    sx = width / input_width;
+    sy = height / input_height;
   } else if(rwidth != R_NilValue){
     width = Rf_asInteger(rwidth);
-    sx = sy = width / dimensions.width;
-    height = round(dimensions.height * sy);
+    sx = sy = width / input_width;
+    height = round(input_height * sy);
   } else {
     height = Rf_asInteger(rheight);
-    sx = sy = height / dimensions.height;
-    width = round(dimensions.width * sx);
+    sx = sy = height / input_height;
+    width = round(input_width * sx);
   }
   switch(Rf_asInteger(format)){
   case 0:
@@ -139,11 +224,15 @@ SEXP R_rsvg(SEXP data, SEXP rwidth, SEXP rheight, SEXP format, SEXP css){
   case 1:
     return write_png(svg, width, height, sx, sy);
   case 2:
-    return write_stream(svg, width, height, sx, sy, cairo_pdf_surface_create_for_stream);
+    return write_stream(svg, width, height, sx, sy, cairo_pdf_surface_create_for_stream, FALSE);
   case 3:
-    return write_stream(svg, width, height, sx, sy, cairo_svg_surface_create_for_stream);
+    return write_stream(svg, width, height, sx, sy, cairo_svg_surface_create_for_stream, FALSE);
   case 4:
-    return write_stream(svg, width, height, sx, sy, cairo_ps_surface_create_for_stream);
+    return write_stream(svg, width, height, sx, sy, cairo_ps_surface_create_for_stream, FALSE);
+  case 5:
+    return write_stream(svg, width, height, sx, sy, cairo_ps_surface_create_for_stream, TRUE);
+  case 6:
+    return write_native_raster(svg, width, height, sx, sy);
   }
   return R_NilValue;
 }
